# -*- coding: utf-8 -*-
# Generated by Django 1.11.23 on 2020-01-15 02:55
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("engine", "0021_auto_20191213_0725"),
    ]

    operations = [
        migrations.AddField(
            model_name="scheduleservice",
            name="multi_callback_enabled",
            field=models.BooleanField(default=False, verbose_name="是否支持多次回调"),
        ),
    ]



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from django.test import TestCase

from pipeline.contrib.external_plugins.tests.mock import *  # noqa
from pipeline.contrib.external_plugins.tests.mock_settings import *  # noqa
from pipeline.contrib.external_plugins.utils.importer.fs import FSModuleImporter

GET_FILE_RETURN = "GET_FILE_RETURN"
GET_SOURCE_RETURN = "a=1"
IS_PACKAGE_RETURN = True
_FETCH_FILE_RETURN = "_FETCH_FILE_RETURN"


class FSModuleImporterTestCase(TestCase):
    def setUp(self):
        self.path = "/usr/imp/custom_components/"
        self.path_without_salsh = "/usr/imp/custom_components"
        self.fullname = "module1.module2.module3"
        self.module_url = "/usr/imp/custom_components/module1/module2/module3.py"
        self.package_url = "/usr/imp/custom_components/module1/module2/module3/__init__.py"

    def test__init__(self):
        importer = FSModuleImporter(name="name", modules=[], path=self.path)
        self.assertEqual(self.path, importer.path)

        importer = FSModuleImporter(name="name", modules=[], path=self.path_without_salsh)
        self.assertEqual(self.path, importer.path)

    def test_is_package(self):
        importer = FSModuleImporter(name="name", modules=[], path=self.path)

        with patch(OS_PATH_EXISTS, MagicMock(return_value=True)):
            self.assertTrue(importer.is_package(self.fullname))

        with patch(OS_PATH_EXISTS, MagicMock(return_value=False)):
            self.assertFalse(importer.is_package(self.fullname))

    @patch(UTILS_IMPORTER_FS_GET_FILE, MagicMock(return_value=GET_FILE_RETURN))
    @patch(UTILS_IMPORTER_FS_GET_SOURCE, MagicMock(return_value=GET_SOURCE_RETURN))
    def test_get_code(self):
        expect_code = compile(GET_SOURCE_RETURN, GET_FILE_RETURN, "exec")
        importer = FSModuleImporter(name="name", modules=[], path=self.path)

        self.assertEqual(expect_code, importer.get_code(self.fullname))

    @patch(UTILS_IMPORTER_FS_IS_PACKAGE, MagicMock(return_value=IS_PACKAGE_RETURN))
    @patch(UTILS_IMPORTER_FS__FETCH_FILE_CONTENT, MagicMock(return_value=_FETCH_FILE_RETURN))
    def test_get_source(self):
        importer = FSModuleImporter(name="name", modules=[], path=self.path)

        self.assertEqual(_FETCH_FILE_RETURN, importer.get_source(self.fullname))
        importer._fetch_file_content.assert_called_once_with(
            importer._file_path(self.fullname, is_pkg=IS_PACKAGE_RETURN)
        )

    @patch(UTILS_IMPORTER_FS_IS_PACKAGE, MagicMock(return_value=IS_PACKAGE_RETURN))
    @patch(UTILS_IMPORTER_FS__FETCH_FILE_CONTENT, MagicMock(return_value=None))
    def test_get_source__fetch_none(self):
        importer = FSModuleImporter(name="name", modules=[], path=self.path)

        self.assertRaises(ImportError, importer.get_source, self.fullname)
        importer._fetch_file_content.assert_called_once_with(
            importer._file_path(self.fullname, is_pkg=IS_PACKAGE_RETURN)
        )

    def test_get_path(self):
        importer = FSModuleImporter(name="name", modules=[], path=self.path)

        self.assertEqual(importer.get_path(self.fullname), ["/usr/imp/custom_components/module1/module2/module3"])

    def test_get_file(self):
        importer = FSModuleImporter(name="name", modules=[], path=self.path)

        with patch(UTILS_IMPORTER_FS_IS_PACKAGE, MagicMock(return_value=True)):
            self.assertEqual(importer.get_file(self.fullname), self.package_url)

        with patch(UTILS_IMPORTER_FS_IS_PACKAGE, MagicMock(return_value=False)):
            self.assertEqual(importer.get_file(self.fullname), self.module_url)

    def test__file_path(self):
        importer = FSModuleImporter(name="name", modules=[], path=self.path)

        self.assertEqual(importer._file_path(self.fullname, is_pkg=True), self.package_url)
        self.assertEqual(importer._file_path(self.fullname, is_pkg=False), self.module_url)

    def test__fetch_file__nocache(self):
        importer = FSModuleImporter(name="name", modules=[], path=self.path, use_cache=False)

        first_file_content = "first_file_content"
        second_file_content = "second_file_content"

        with patch(UTILS_IMPORTER_FS__GET_FILE_CONTENT, MagicMock(return_value=first_file_content)):
            self.assertEqual(importer._fetch_file_content(self.module_url), first_file_content)
            self.assertEqual(importer.file_cache, {})

        with patch(UTILS_IMPORTER_FS__GET_FILE_CONTENT, MagicMock(return_value=second_file_content)):
            self.assertEqual(importer._fetch_file_content(self.module_url), second_file_content)
            self.assertEqual(importer.file_cache, {})

        with patch(UTILS_IMPORTER_FS__GET_FILE_CONTENT, MagicMock(return_value=None)):
            self.assertIsNone(importer._fetch_file_content(self.module_url))
            self.assertEqual(importer.file_cache, {})

    def test__fetch_file__use_cache(self):
        importer = FSModuleImporter(name="name", modules=[], path=self.path)

        first_file_content = "first_file_content"
        second_file_content = "second_file_content"

        with patch(UTILS_IMPORTER_FS__GET_FILE_CONTENT, MagicMock(return_value=first_file_content)):
            self.assertEqual(importer._fetch_file_content(self.module_url), first_file_content)
            self.assertEqual(importer.file_cache[self.module_url], first_file_content)

        with patch(UTILS_IMPORTER_FS__GET_FILE_CONTENT, MagicMock(return_value=second_file_content)):
            self.assertEqual(importer._fetch_file_content(self.module_url), first_file_content)
            self.assertEqual(importer.file_cache[self.module_url], first_file_content)

        with patch(UTILS_IMPORTER_FS__GET_FILE_CONTENT, MagicMock(return_value=None)):
            self.assertIsNone(importer._fetch_file_content(self.package_url))
            self.assertEqual(importer.file_cache[self.package_url], None)

        with patch(UTILS_IMPORTER_FS__GET_FILE_CONTENT, MagicMock(return_value=second_file_content)):
            self.assertIsNone(importer._fetch_file_content(self.package_url))
            self.assertEqual(importer.file_cache[self.package_url], None)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

ITERATED = 1
NEW = 0
ITERATING = -1


def has_circle(graph):
    # init marks
    marks = {}
    for node in graph:
        # marks as not iterated
        marks[node] = NEW

    # dfs every node
    for cur_node in graph:
        trace = [cur_node]
        for node in graph[cur_node]:
            if marks[node] == ITERATED:
                continue
            trace.append(node)
            # return immediately when circle be detected
            if _has_circle(graph, node, marks, trace):
                return True, trace
            trace.pop()
        # mark as iterated
        marks[cur_node] = ITERATED

    return False, []


def _has_circle(graph, cur_node, marks, trace):
    # detect circle when iterate to a node which been marked as -1
    if marks[cur_node] == ITERATING:
        return True
    # mark as iterating
    marks[cur_node] = ITERATING
    # dfs
    for node in graph[cur_node]:
        # return immediately when circle be detected
        trace.append(node)
        if _has_circle(graph, node, marks, trace):
            return True
        trace.pop()
    # mark as iterated
    marks[cur_node] = ITERATED

    return False


def convert_bytes_to_str(obj):

    converted = set()

    def _convert(obj, converted):
        if isinstance(obj, dict):
            new_dict = obj.__class__()

            for attr, value in obj.items():

                if isinstance(attr, bytes):
                    attr = attr.decode("utf-8")

                value = _convert(value, converted)

                new_dict[attr] = value

            obj = new_dict

        if isinstance(obj, list):
            new_list = obj.__class__()

            for item in obj:
                new_list.append(_convert(item, converted))

            obj = new_list

        elif isinstance(obj, bytes):

            try:
                obj = obj.decode("utf-8")
            except Exception:
                pass

        elif hasattr(obj, "__dict__"):

            if id(obj) in converted:
                return obj
            else:
                converted.add(id(obj))

            new__dict__ = {}

            for attr, value in obj.__dict__.items():

                if isinstance(attr, bytes):
                    attr = attr.decode("utf-8")

                new__dict__[attr] = _convert(value, converted)

            obj.__dict__ = new__dict__

        return obj

    return _convert(obj, converted)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

# mock str return value of Built-in Functions，make str(func) return "func" rather than "<built-in function func>"

import builtins
import importlib

from pipeline.conf import default_settings

SANDBOX = {}


class MockStrMeta(type):
    def __new__(cls, name, bases, attrs):
        new_cls = super(MockStrMeta, cls).__new__(cls, name, bases, attrs)
        SANDBOX.update({new_cls.str_return: new_cls})
        return new_cls

    def __str__(cls):
        return cls.str_return

    def __call__(cls, *args, **kwargs):
        return cls.call(*args, **kwargs)


def _shield_words(sandbox, words):
    for shield_word in words:
        sandbox[shield_word] = None


class ModuleObject:
    def __init__(self, sub_paths, module):
        if len(sub_paths) == 1:
            setattr(self, sub_paths[0], module)
            return
        setattr(self, sub_paths[0], ModuleObject(sub_paths[1:], module))


def _import_modules(sandbox, modules):
    for mod_path, alias in modules.items():
        mod = importlib.import_module(mod_path)
        sub_paths = alias.split(".")
        if len(sub_paths) == 1:
            sandbox[alias] = mod
        else:
            sandbox[sub_paths[0]] = ModuleObject(sub_paths[1:], mod)


def _mock_builtins():
    """
    @summary: generate mock class of built-in functions like id,int
    """
    for func_name in dir(builtins):
        if func_name.lower() == func_name and not func_name.startswith("_"):
            new_func_name = "Mock{}".format(func_name.capitalize())
            MockStrMeta(new_func_name, (object,), {"call": getattr(builtins, func_name), "str_return": func_name})


_mock_builtins()

_shield_words(SANDBOX, default_settings.MAKO_SANDBOX_SHIELD_WORDS)

_import_modules(SANDBOX, default_settings.MAKO_SANDBOX_IMPORT_MODULES)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import ujson as json
from django.db import models
from django.utils.translation import ugettext_lazy as _

from pipeline.constants import PIPELINE_DEFAULT_PRIORITY
from pipeline.contrib.periodic_task.signals import periodic_task_start_failed
from pipeline.exceptions import InvalidOperationException
from pipeline.models import (
    CompressJSONField,
    PipelineInstance,
    PipelineTemplate,
    Snapshot,
)
from pipeline.utils.uniqid import uniqid
from django_celery_beat.models import (
    PeriodicTask as DjangoCeleryBeatPeriodicTask,
    CrontabSchedule as DjangoCeleryBeatCrontabSchedule,
)

from pipeline.contrib.periodic_task.djcelery.models import *  # noqa

BAMBOO_ENGINE_TRIGGER_TASK = "pipeline.contrib.periodic_task.tasks.bamboo_engine_periodic_task_start"


class PeriodicTaskManager(models.Manager):
    def create_task(
        self,
        name,
        template,
        cron,
        data,
        creator,
        timezone=None,
        extra_info=None,
        spread=False,
        priority=PIPELINE_DEFAULT_PRIORITY,
        queue="",
        trigger_task="",
    ):
        snapshot = Snapshot.objects.create_snapshot(data)
        schedule, _ = DjangoCeleryBeatCrontabSchedule.objects.get_or_create(
            minute=cron.get("minute", "*"),
            hour=cron.get("hour", "*"),
            day_of_week=cron.get("day_of_week", "*"),
            day_of_month=cron.get("day_of_month", "*"),
            month_of_year=cron.get("month_of_year", "*"),
            timezone=timezone or "UTC",
        )
        _ = schedule.schedule  # noqa

        task = self.create(
            name=name,
            template=template,
            snapshot=snapshot,
            cron=schedule.__str__(),
            creator=creator,
            extra_info=extra_info,
            priority=priority,
            queue=queue,
        )

        kwargs = {"period_task_id": task.id, "spread": spread}
        celery_task = DjangoCeleryBeatPeriodicTask.objects.create(
            crontab=schedule,
            name=uniqid(),
            task=trigger_task or "pipeline.contrib.periodic_task.tasks.periodic_task_start",
            enabled=False,
            kwargs=json.dumps(kwargs),
        )
        task.celery_task = celery_task
        task.save()
        return task


class PeriodicTask(models.Model):
    name = models.CharField(_("周期任务名称"), max_length=64)
    template = models.ForeignKey(
        PipelineTemplate,
        related_name="periodic_tasks",
        to_field="template_id",
        verbose_name=_("周期任务对应的模板"),
        null=True,
        on_delete=models.deletion.SET_NULL,
    )
    cron = models.CharField(_("调度策略"), max_length=128)
    celery_task = models.ForeignKey(
        DjangoCeleryBeatPeriodicTask, verbose_name=_("celery 周期任务实例"), null=True, on_delete=models.SET_NULL,
    )
    snapshot = models.ForeignKey(
        Snapshot, related_name="periodic_tasks", verbose_name=_("用于创建流程实例的结构数据"), on_delete=models.DO_NOTHING,
    )
    total_run_count = models.PositiveIntegerField(_("执行次数"), default=0)
    last_run_at = models.DateTimeField(_("上次运行时间"), null=True)
    creator = models.CharField(_("创建者"), max_length=32, default="")
    priority = models.IntegerField(_("流程优先级"), default=PIPELINE_DEFAULT_PRIORITY)
    queue = models.CharField(_("流程使用的队列名"), max_length=512, default="")
    extra_info = CompressJSONField(verbose_name=_("额外信息"), null=True)

    objects = PeriodicTaskManager()

    def __unicode__(self):
        return "{name}({id})".format(name=self.name, id=self.id)

    @property
    def enabled(self):
        return self.celery_task.enabled

    @property
    def execution_data(self):
        return self.snapshot.data

    @property
    def form(self):
        form = {
            key: var_info
            for key, var_info in list(self.execution_data["constants"].items())
            if var_info["show_type"] == "show"
        }
        return form

    def delete(self, using=None):
        self.set_enabled(False)
        self.celery_task.delete()
        PeriodicTaskHistory.objects.filter(periodic_task=self).delete()
        return super(PeriodicTask, self).delete(using)

    def set_enabled(self, enabled):
        self.celery_task.enabled = enabled
        self.celery_task.save()

    def modify_cron(self, cron, timezone=None):
        if self.enabled:
            raise InvalidOperationException("can not modify cron when task is enabled")
        schedule, _ = DjangoCeleryBeatCrontabSchedule.objects.get_or_create(
            minute=cron.get("minute", "*"),
            hour=cron.get("hour", "*"),
            day_of_week=cron.get("day_of_week", "*"),
            day_of_month=cron.get("day_of_month", "*"),
            month_of_year=cron.get("month_of_year", "*"),
            timezone=timezone or "UTC",
        )
        # try to initiate schedule object
        _ = schedule.schedule  # noqa
        self.cron = schedule.__str__()
        self.celery_task.crontab = schedule
        self.celery_task.save()
        self.save()

    def modify_constants(self, constants):
        if self.enabled:
            raise InvalidOperationException("can not modify constants when task is enabled")
        exec_data = self.execution_data
        for key, value in list(constants.items()):
            if key in exec_data["constants"]:
                exec_data["constants"][key]["value"] = value
        self.snapshot.data = exec_data
        self.snapshot.save()
        return exec_data["constants"]


class PeriodicTaskHistoryManager(models.Manager):
    def record_schedule(self, periodic_task, pipeline_instance, ex_data, start_success=True):
        history = self.create(
            periodic_task=periodic_task,
            pipeline_instance=pipeline_instance,
            ex_data=ex_data,
            start_success=start_success,
            priority=periodic_task.priority,
            queue=periodic_task.queue,
        )

        if not start_success:
            periodic_task_start_failed.send(sender=PeriodicTask, periodic_task=periodic_task, history=history)

        return history


class PeriodicTaskHistory(models.Model):
    periodic_task = models.ForeignKey(
        PeriodicTask, related_name="instance_rel", verbose_name=_("周期任务"), null=True, on_delete=models.DO_NOTHING,
    )
    pipeline_instance = models.ForeignKey(
        PipelineInstance,
        related_name="periodic_task_rel",
        verbose_name=_("Pipeline 实例"),
        to_field="instance_id",
        null=True,
        on_delete=models.DO_NOTHING,
    )
    ex_data = models.TextField(_("异常信息"))
    start_at = models.DateTimeField(_("开始时间"), auto_now_add=True)
    start_success = models.BooleanField(_("是否启动成功"), default=True)
    priority = models.IntegerField(_("流程优先级"), default=PIPELINE_DEFAULT_PRIORITY)
    queue = models.CharField(_("流程使用的队列名"), max_length=512, default="")

    objects = PeriodicTaskHistoryManager()



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from queue import Queue

from pipeline.core.flow.activity import Activity
from pipeline.core.flow.gateway import Gateway
from pipeline.exceptions import PipelineException


class PipelineSpec(object):
    def __init__(self, start_event, end_event, flows, activities, gateways, data, context):
        objects = {start_event.id: start_event, end_event.id: end_event}
        for act in activities:
            objects[act.id] = act
        for gw in gateways:
            objects[gw.id] = gw

        self.start_event = start_event
        self.end_event = end_event
        self.flows = flows
        self.activities = activities
        self.gateways = gateways
        self.data = data
        self.objects = objects
        self.context = context

    def prune(self, keep_from, keep_to):
        if keep_from != self.start_event.id:
            self.start_event.outgoing = None

        if keep_to != self.end_event.id:
            self.end_event.incoming = None

        self.activities = []
        self.gateways = []
        self.flows = []

        keep_from_node = self.objects[keep_from]
        keep_to_node = self.objects[keep_to]

        keep_from_node.incoming = None
        keep_to_node.outgoing = None

        to_be_process = Queue()
        to_be_process.put(keep_from_node)

        new_objects = {}
        keep_to_incoming_flows = []

        while not to_be_process.empty():
            node = to_be_process.get()

            if issubclass(node.__class__, Activity):
                self.activities.append(node)
            elif issubclass(node.__class__, Gateway):
                self.gateways.append(node)

            new_objects[node.id] = node

            if node.id == keep_to_node.id:
                continue

            for out in node.outgoing:

                self.flows.append(out)

                if out.target.id not in new_objects:
                    next_node = out.target
                    if next_node.id == keep_to_node.id:
                        keep_to_incoming_flows.append(out)
                    to_be_process.put(next_node)

        keep_to_node.incoming.flows = keep_to_incoming_flows
        keep_to_node.incoming.flow_dict = {}
        for flow in keep_to_incoming_flows:
            keep_to_node.incoming.flow_dict[flow.id] = flow

        self.objects = new_objects


class PipelineShell(object):
    def __init__(self, id, data):
        self.id = id
        self.data = data

    def shell(self):
        return PipelineShell(id=self.id, data=self.data)


class Pipeline(object):
    def __init__(self, id, pipeline_spec, parent=None):
        self.id = id
        self.spec = pipeline_spec
        self.parent = parent

    @property
    def data(self):
        return self.spec.data

    @property
    def context(self):
        return self.spec.context

    @property
    def start_event(self):
        return self.spec.start_event

    @property
    def end_event(self):
        return self.spec.end_event

    @property
    def all_nodes(self):
        return self.spec.objects

    def data_for_node(self, node):
        node = self.spec.objects.get(node.id)
        if not node:
            raise PipelineException("Can not find node %s in this pipeline." % node.id)
        return node.data

    def node(self, id):
        return self.spec.objects.get(id)

    def prune(self, keep_from, keep_to):
        self.spec.prune(keep_from=keep_from, keep_to=keep_to)

    def shell(self):
        return PipelineShell(id=self.id, data=self.data)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from .activity import SubProcess  # noqa
from .activity import AbstractIntervalGenerator  # noqa

from .activity import (  # noqa
    DefaultIntervalGenerator,
    LinearIntervalGenerator,
    NullIntervalGenerator,
    Service,
    ServiceActivity,
    SquareIntervalGenerator,
    StaticIntervalGenerator,
)
from .base import SequenceFlow  # noqa
from .event import (  # noqa
    EmptyEndEvent,
    EmptyStartEvent,
    EndEvent,
    ExecutableEndEvent,
    StartEvent,
)
from .gateway import (  # noqa
    Condition,
    ConditionalParallelGateway,
    ConvergeGateway,
    ExclusiveGateway,
    ParallelGateway,
)
from .signals import post_new_end_event_register


class FlowNodeClsFactory(object):
    nodes_cls = {
        ServiceActivity.__name__: ServiceActivity,
        SubProcess.__name__: SubProcess,
        EmptyEndEvent.__name__: EmptyEndEvent,
        EmptyStartEvent.__name__: EmptyStartEvent,
        ParallelGateway.__name__: ParallelGateway,
        ConditionalParallelGateway.__name__: ConditionalParallelGateway,
        ExclusiveGateway.__name__: ExclusiveGateway,
        ConvergeGateway.__name__: ConvergeGateway,
    }

    @classmethod
    def _nodes_types_filter(cls, cls_filter):
        types = []
        for node_type, node_cls in list(cls.nodes_cls.items()):
            if not cls_filter(node_cls):
                types.append(node_type)

        return types

    @classmethod
    def node_types_without_start_event(cls):
        return cls._nodes_types_filter(cls_filter=lambda node_cls: issubclass(node_cls, StartEvent))

    @classmethod
    def node_types_without_start_end_event(cls):
        return cls._nodes_types_filter(
            cls_filter=lambda node_cls: issubclass(node_cls, EndEvent) or issubclass(node_cls, StartEvent)
        )

    @classmethod
    def get_node_cls(cls, key):
        return cls.nodes_cls.get(key)

    @classmethod
    def register_node(cls, key, node_cls):
        if key in cls.nodes_cls:
            raise KeyError("node with key({key}) is already exist: {node}".format(key=key, node=cls.nodes_cls[key]))

        cls.nodes_cls[key] = node_cls

        if issubclass(node_cls, EndEvent):
            post_new_end_event_register.send(sender=EndEvent, node_type=key, node_cls=node_cls)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.23 on 2019-12-13 08:28
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("periodic_task", "0003_auto_20191213_0819"),
    ]

    operations = [
        migrations.AddField(
            model_name="periodictaskhistory",
            name="priority",
            field=models.IntegerField(default=100, verbose_name="流程优先级"),
        ),
        migrations.AddField(
            model_name="periodictaskhistory",
            name="queue",
            field=models.CharField(default="", max_length=512, verbose_name="流程使用的队列名"),
        ),
    ]



# Generated by Django 2.2.19 on 2021-06-07 09:23

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("pipeline", "0028_auto_20201227_1952"),
    ]

    operations = [
        migrations.AddField(
            model_name="templaterelationship",
            name="always_use_latest",
            field=models.BooleanField(default=False, verbose_name="是否永远使用最新版本"),
        ),
    ]



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""


class PipelineException(Exception):
    pass


class FlowTypeError(PipelineException):
    pass


class InvalidOperationException(PipelineException):
    pass


class ConditionExhaustedException(PipelineException):
    pass


class EvaluationException(PipelineException):
    pass


class NodeNotExistException(PipelineException):
    pass


class SourceKeyException(NodeNotExistException):
    pass


class VariableHydrateException(PipelineException):
    pass


class ParserException(PipelineException):
    pass


class SubprocessRefError(PipelineException):
    pass


class TemplateImportError(PipelineException):
    pass


class SubprocessExpiredError(PipelineException):
    pass


#
# data exception
#


class DataException(PipelineException):
    pass


class DataInitException(DataException):
    pass


class DataAttrException(DataException):
    pass


class DataTypeErrorException(DataException):
    pass


class CycleErrorException(DataException):
    pass


class ConnectionValidateError(DataException):
    def __init__(self, failed_nodes, detail, *args):
        self.failed_nodes = failed_nodes
        self.detail = detail
        super(ConnectionValidateError, self).__init__(*args)


class ConvergeMatchError(DataException):
    def __init__(self, gateway_id, *args):
        self.gateway_id = gateway_id
        super(ConvergeMatchError, self).__init__(*args)


class StreamValidateError(DataException):
    def __init__(self, node_id, *args):
        self.node_id = node_id
        super(StreamValidateError, self).__init__(*args)


class IsolateNodeError(DataException):
    pass


#
# component exception
#


class ComponentException(PipelineException):
    pass


class ComponentDataFormatException(ComponentException):
    pass


class ComponentNotExistException(ComponentException):
    pass


class ComponentDataLackException(ComponentDataFormatException):
    pass


#
# tag exception
#


class PipelineError(Exception):
    pass


class TagError(PipelineError):
    pass


class AttributeMissingError(TagError):
    pass


class AttributeValidationError(TagError):
    pass


#
# constant exception
#
class ConstantException(PipelineException):
    pass


class ConstantNotExistException(ConstantException):
    pass


class ConstantReferenceException(ConstantException):
    pass


class ConstantTypeException(ConstantException):
    pass


class ConstantSyntaxException(ConstantException):
    pass


#
# context exception
#
class ContextError(PipelineError):
    pass


class ReferenceNotExistError(ContextError):
    pass


class InsufficientVariableError(ContextError):
    pass


#
# periodic task exception
#
class InvalidCrontabException(PipelineException):
    pass



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import imp
import logging
import sys
import traceback
from abc import ABCMeta, abstractmethod
from contextlib import contextmanager

from pipeline.contrib.external_plugins.utils import requirement

logger = logging.getLogger("root")


@contextmanager
def hook_sandbox(hook, fullname):
    hook_name = hook.__func__.__name__
    try:
        logger.info("Execute {hook_name} for {module}".format(module=fullname, hook_name=hook_name))
        yield
    except Exception:
        logger.error(
            "{module} {hook_name} raise exception: {traceback}".format(
                module=fullname, hook_name=hook_name, traceback=traceback.format_exc()
            )
        )


class NonstandardModuleImporter(object, metaclass=ABCMeta):
    def __init__(self, modules, name=None):
        self.name = name
        self.modules = modules

    def find_module(self, fullname, path=None):
        logger.info("=============FINDER: {cls}".format(cls=self.__class__.__name__))
        logger.info("Try to find module: {module} in path: {path}".format(module=fullname, path=path))

        logger.info("Check if in declared nonstandard modules: {modules}".format(modules=self.modules))
        root_parent = fullname.split(".")[0]
        if root_parent not in self.modules:
            logger.info("Root module({module}) are not find in nonstandard modules".format(module=root_parent))
            return None

        logger.info("Check if is built-in module")
        try:
            loader = imp.find_module(fullname, path)
            if loader:
                logger.info("Found {module} locally".format(module=fullname))
                return None
        except ImportError:
            pass

        logger.info("Checking if is name repetition")
        if fullname.split(".").count(fullname.split(".")[-1]) > 1:
            logger.info("Found {module} locally".format(module=fullname))
            return None

        with hook_sandbox(fullname=fullname, hook=self.accept_find_module_request_hook):
            self.accept_find_module_request_hook(fullname=fullname, path=path)

        return self

    def load_module(self, fullname):
        try:
            imp.acquire_lock()

            logger.info("=============LOADER: {cls}".format(cls=self.__class__.__name__))
            logger.info("Try to load module: {module}".format(module=fullname))

            if fullname in sys.modules:
                logger.info("Module {module} already loaded".format(module=fullname))
                return sys.modules[fullname]

            is_pkg = self.is_package(fullname)

            try:
                src_code = self.get_source(fullname)
            except ImportError as e:
                logger.info("Get source code for {module} error: {message}".format(module=fullname, message=e))
                return None

            logger.info("Importing {module}".format(module=fullname))
            mod = sys.modules.setdefault(fullname, imp.new_module(fullname))

            with hook_sandbox(fullname=fullname, hook=self.pre_load_module_hook):
                self.pre_load_module_hook(fullname=fullname, module=mod)

            mod.__file__ = self.get_file(fullname)
            mod.__loader__ = self
            mod.__name__ = fullname
            if is_pkg:
                mod.__path__ = self.get_path(fullname)
                mod.__package__ = fullname
            else:
                mod.__package__ = fullname.rpartition(".")[0]

            logger.info("Module prepared, ready to execute source code for {module}".format(module=fullname))
            logger.info("Source code for {module}:\n{src_code}".format(module=fullname, src_code=src_code))

            self._execute_src_code(src_code=src_code, module=mod)

            with hook_sandbox(fullname=fullname, hook=self.post_load_module_hook):
                self.post_load_module_hook(fullname=fullname, module=mod)

            return mod

        except Exception:

            with hook_sandbox(fullname=fullname, hook=self.import_error_hook):
                self.import_error_hook(fullname)

            err_msg = "{module} import raise exception: {traceback}".format(
                module=fullname, traceback=traceback.format_exc()
            )
            logger.error(err_msg)

            if fullname in sys.modules:
                logger.info("Remove module {module} from sys.modules".format(module=fullname))
                del sys.modules[fullname]

            raise ImportError(err_msg)

        finally:
            imp.release_lock()

    def _execute_src_code(self, src_code, module):
        exec(src_code, module.__dict__)

    @abstractmethod
    def is_package(self, fullname):
        raise NotImplementedError()

    @abstractmethod
    def get_code(self, fullname):
        raise NotImplementedError()

    @abstractmethod
    def get_source(self, fullname):
        raise NotImplementedError()

    @abstractmethod
    def get_file(self, fullname):
        return NotImplementedError()

    @abstractmethod
    def get_path(self, fullname):
        return NotImplementedError()

    def accept_find_module_request_hook(self, fullname, path):
        pass

    def pre_load_module_hook(self, fullname, module):
        pass

    def post_load_module_hook(self, fullname, module):
        pass

    def import_error_hook(self, fullname):
        pass


class AutoInstallRequirementsImporter(NonstandardModuleImporter, metaclass=ABCMeta):
    def post_load_module_hook(self, fullname, module):
        requirements = getattr(module, "__requirements__", [])
        if not isinstance(requirements, list) or not requirements:
            return

        sys.stdout.write(
            "Start to install requirements({reqs}) for module({mod})\n".format(
                reqs=",".join(requirements), mod=fullname
            )
        )
        requirement.install(requirements)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from copy import deepcopy

from django.test import TestCase

from pipeline.contrib.external_plugins import exceptions
from pipeline.contrib.external_plugins.models import ExternalPackageSource, GitRepoSource, source_cls_factory
from pipeline.contrib.external_plugins.tests.mock import *  # noqa
from pipeline.contrib.external_plugins.tests.mock_settings import *  # noqa

SOURCE_NAME = "source_name"
PACKAGES = {
    "root_package_1": {"version": "", "modules": ["test1", "test2"]},
    "root_package_2": {"version": "", "modules": ["test3", "test4"]},
    "root_package_3": {"version": "", "modules": ["test5", "test6"]},
}
FROM_CONFIG = True
REPO_RAW_ADDRESS = "REPO_RAW_ADDRESS"
BRANCH = "master"

OLD_SOURCE_1 = {
    "name": "source_1",
    "details": {"repo_raw_address": "old_address", "branch": "stage"},
    "packages": {"root_package": {"version": "", "modules": ["test1"]}},
}

OLD_SOURCE_3 = {
    "name": "source_3",
    "details": {"repo_raw_address": "old_address_3", "branch": "master"},
    "packages": {"root_package": {"version": "", "modules": ["test5"]}},
}

SOURCE_1 = {
    "name": "source_1",
    "details": {"repo_raw_address": "https://github.com/homholueng/plugins_example_1", "branch": "master"},
    "packages": {"root_package": {"version": "", "modules": ["test1", "test2"]}},
}

SOURCE_2 = {
    "name": "source_2",
    "details": {"repo_raw_address": "https://github.com/homholueng/plugins_example_2", "branch": "master"},
    "packages": {"root_package": {"version": "", "modules": ["test3", "test4"]}},
}

SOURCE_4 = {
    "name": "source_4",
    "details": {"repo_raw_address": "https://github.com/homholueng/plugins_example_4", "branch": "master"},
    "packages": {"root_package": {"version": "", "modules": ["test5", "test6"]}},
}

GIT_SOURCE_CONFIGS = [SOURCE_1, SOURCE_2, SOURCE_4]


class ExternalPackageSourceTestCase(TestCase):
    def setUp(self):
        GitRepoSource.objects.create_source(
            name=SOURCE_NAME,
            packages=PACKAGES,
            from_config=FROM_CONFIG,
            repo_raw_address=REPO_RAW_ADDRESS,
            branch=BRANCH,
        )

    def tearDown(self):
        GitRepoSource.objects.all().delete()

    def test_create_source(self):
        source_1 = GitRepoSource.objects.get(name=SOURCE_NAME)
        self.assertEqual(source_1.name, SOURCE_NAME)
        self.assertEqual(source_1.packages, PACKAGES)
        self.assertEqual(source_1.from_config, FROM_CONFIG)
        self.assertEqual(source_1.repo_raw_address, REPO_RAW_ADDRESS)
        self.assertEqual(source_1.branch, BRANCH)

    def test_remove_source(self):
        source_1 = GitRepoSource.objects.get(name=SOURCE_NAME)

        self.assertRaises(exceptions.InvalidOperationException, GitRepoSource.objects.remove_source, source_1.id)

        source_1.from_config = False
        source_1.save()

        GitRepoSource.objects.remove_source(source_1.id)

        self.assertFalse(GitRepoSource.objects.filter(id=source_1.id).exists())

    def _assert_source_equals_config(self, source, config):
        self.assertEqual(source.name, config["name"])
        self.assertEqual(source.packages, config["packages"])
        self.assertEqual(source.repo_raw_address, config["details"]["repo_raw_address"])
        self.assertEqual(source.branch, config["details"]["branch"])

    def test_update_source_from_config(self):
        GitRepoSource.objects.all().delete()

        for source in [OLD_SOURCE_1, OLD_SOURCE_3]:
            GitRepoSource.objects.create_source(
                name=source["name"],
                packages=source["packages"],
                from_config=True,
                repo_raw_address=source["details"]["repo_raw_address"],
                branch=source["details"]["branch"],
            )

        GitRepoSource.objects.update_source_from_config(GIT_SOURCE_CONFIGS)

        self.assertFalse(GitRepoSource.objects.filter(name=OLD_SOURCE_3["name"]).exists())

        for config in GIT_SOURCE_CONFIGS:
            source = GitRepoSource.objects.get(name=config["name"])
            self.assertTrue(source.from_config)
            self._assert_source_equals_config(source, config)

    def test_modules(self):
        source = GitRepoSource.objects.get(name=SOURCE_NAME)

        modules = []
        for package_info in list(PACKAGES.values()):
            modules.extend(package_info["modules"])

        self.assertEqual(source.modules, modules)

    @patch(MODELS_SOURCE_MANAGER_UPDATE_SOURCE_FROM_CONFIG, MagicMock())
    def test_update_package_source_from_config__empty_configs(self):
        ExternalPackageSource.update_package_source_from_config([])
        for source_model_cls in list(source_cls_factory.values()):
            source_model_cls.objects.update_source_from_config.assert_called_with(configs=[])

    @patch(MODELS_SOURCE_MANAGER_UPDATE_SOURCE_FROM_CONFIG, MagicMock())
    def test_update_package_source_from_config__normal_case(self):
        source_configs = [
            {"name": "1", "type": "git"},
            {"name": "2", "type": "git"},
            {"name": "3", "type": "s3"},
            {"name": "4", "type": "fs"},
        ]
        ExternalPackageSource.update_package_source_from_config(source_configs)
        GitRepoSource.objects.update_source_from_config.assert_has_calls(
            [call(configs=[{"name": "1"}, {"name": "2"}]), call(configs=[{"name": "3"}]), call(configs=[{"name": "4"}])]
        )

    def test_update_package_source_from_config__unsupported_source_type(self):
        source_configs = [{"name": "1", "type": "wrong_type"}]
        self.assertRaises(KeyError, ExternalPackageSource.update_package_source_from_config, source_configs)

    def test_update_source_from_config__name_conflict(self):
        source = deepcopy(SOURCE_1)
        source["type"] = "git"
        ExternalPackageSource.update_package_source_from_config([source])
        GitRepoSource.objects.filter(name=source["name"]).update(from_config=False)
        self.assertRaises(
            exceptions.InvalidOperationException, ExternalPackageSource.update_package_source_from_config, [source]
        )



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from django.test import TestCase

from pipeline.django_signal_valve import valve
from pipeline.django_signal_valve.models import Signal
from pipeline.django_signal_valve.tests import mock_signal


class TestValve(TestCase):
    def setUp(self):
        valve.unload_valve_function()

    def test_set_valve_function(self):
        self.assertRaises(Exception, valve.set_valve_function, args=[1])

        def func():
            return True

        valve.unload_valve_function()
        valve.set_valve_function(func)
        self.assertEqual(valve.valve_function(), func)
        self.assertRaises(Exception, valve.set_valve_function, args=[func])

        valve.__valve_function = None

    def test_send_on_valve_is_none(self):
        kwargs_1 = {"1": 1}
        kwargs_2 = {"2": 2}

        valve.unload_valve_function()
        valve.send(mock_signal, "signal_1", **kwargs_1)
        valve.send(mock_signal, "signal_1", **kwargs_2)
        self.assertEqual(mock_signal.signal_1.history[0], kwargs_1)
        self.assertEqual(mock_signal.signal_1.history[1], kwargs_2)

        mock_signal.clear()

    def test_send_on_valve_opened(self):
        kwargs_1 = {"1": 1}
        kwargs_2 = {"2": 2}

        def is_valve_closed():
            return False

        valve.unload_valve_function()
        valve.set_valve_function(is_valve_closed)
        valve.send(mock_signal, "signal_1", **kwargs_1)
        valve.send(mock_signal, "signal_1", **kwargs_2)
        self.assertEqual(mock_signal.signal_1.history[0], kwargs_1)
        self.assertEqual(mock_signal.signal_1.history[1], kwargs_2)

        mock_signal.clear()

    def test_send_on_closed(self):
        kwargs_1 = {"1": 1}
        kwargs_2 = {"2": 2}

        def is_valve_closed():
            return True

        valve.unload_valve_function()
        valve.set_valve_function(is_valve_closed)
        valve.send(mock_signal, "signal_1", **kwargs_1)
        valve.send(mock_signal, "signal_1", **kwargs_2)
        self.assertEqual(len(mock_signal.signal_1.history), 0)

        mock_signal.clear()
        Signal.objects.all().delete()

    def test_open_valve(self):
        kwargs_1 = {"1": 1}
        kwargs_2 = {"2": 2}

        def valve_closed():
            return True

        valve.unload_valve_function()
        valve.set_valve_function(valve_closed)
        valve.send(mock_signal, "signal_1", **kwargs_1)
        valve.send(mock_signal, "signal_1", **kwargs_2)
        self.assertEqual(len(mock_signal.signal_1.history), 0)
        valve.open_valve(mock_signal)
        self.assertEqual(mock_signal.signal_1.history[0], kwargs_1)
        self.assertEqual(mock_signal.signal_1.history[1], kwargs_2)

        mock_signal.clear()



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from copy import deepcopy
from pprint import pformat

from pipeline.exceptions import InvalidOperationException, ReferenceNotExistError


class Context(object):
    def __init__(self, act_outputs, output_key=None, scope=None):
        self.variables = scope or {}
        self.act_outputs = act_outputs
        self._output_key = set(output_key or [])
        self._change_keys = set()
        self._raw_variables = None

    def extract_output(self, activity, set_miss=True):
        self.extract_output_from_data(activity.id, activity.data, set_miss=set_miss)

    def extract_output_from_data(self, activity_id, data, set_miss=True):
        if activity_id in self.act_outputs:
            global_outputs = self.act_outputs[activity_id]
            output = data.get_outputs()
            for key in global_outputs:
                # set value to key if can not find
                # e.g. key: result
                # e.g. global_outputs[key]: result_5hoi2
                if key not in output and not set_miss:
                    continue

                self.variables[global_outputs[key]] = output.get(key, global_outputs[key])
                self.change_keys.add(global_outputs[key])

    def get(self, key):
        try:
            return self.variables[key]
        except KeyError:
            raise ReferenceNotExistError('reference "%s" does not exist.' % key)

    def set_global_var(self, key, val):
        self.variables[key] = val
        self.change_keys.add(key)

    def update_global_var(self, var_dict):
        self.variables.update(var_dict)
        self.change_keys.update(list(var_dict.keys()))

    def mark_as_output(self, key):
        self._output_key.add(key)

    def write_output(self, pipeline):
        from pipeline.core.data import var

        data = pipeline.data
        for key in self._output_key:
            try:
                value = self.get(key)
            except ReferenceNotExistError:
                value = key

            if issubclass(value.__class__, var.Variable):
                value = value.get()
                # break circle
            data.set_outputs(key, value)

    def duplicate_variables(self):
        self._raw_variables = deepcopy(self.variables)

    def clear(self):
        self.variables.clear()
        if self.raw_variables:
            self.raw_variables.clear()

    def recover_variable(self):
        if self.raw_variables is None:
            raise InvalidOperationException("make sure duplicate_variables() is called before do recover")

        # collect all act output key
        act_outputs_keys = set()
        for global_outputs in list(self.act_outputs.values()):
            for output_key in list(global_outputs.values()):
                act_outputs_keys.add(output_key)

        # recover to Variable for which key not in act output
        for key, var in list(self.raw_variables.items()):
            if key not in act_outputs_keys:
                self.variables[key] = deepcopy(var)

    def clear_change_keys(self):
        if hasattr(self, "_change_keys"):
            self.change_keys.clear()

    def sync_change(self, context):
        from pipeline.core.data.var import SpliceVariable

        # sync obvious change keys
        for k in context.change_keys:
            self.set_global_var(k, context.get(k))

        # sync resolved splice value
        for k, child_v in context.variables.items():
            parent_v = self.variables.get(k)
            if isinstance(child_v, SpliceVariable) and isinstance(parent_v, SpliceVariable):
                # if var is resolved in child
                if parent_v._value is None and child_v._value is not None:
                    parent_v._value = child_v._value

    def __repr__(self):
        return "variables:{}\nact_outputs:{}\n_output_key:{}".format(
            pformat(self.variables), pformat(self.act_outputs), pformat(self._output_key)
        )

    def __str__(self):
        return self.__repr__()

    def __unicode__(self):
        return self.__repr__()

    @property
    def change_keys(self):
        if not hasattr(self, "_change_keys"):
            self._change_keys = set()

        return self._change_keys

    @property
    def raw_variables(self):
        if not hasattr(self, "_raw_variables"):
            self._raw_variables = None

        return self._raw_variables


class OutputRef(object):
    def __init__(self, key, context):
        self.key = key
        self.context = context

    @property
    def value(self):
        return self.context.get(self.key)

    def __deepcopy__(self, memodict={}):
        return self



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import logging
import sys

from django.apps import AppConfig
from django.db.utils import InternalError, OperationalError, ProgrammingError

from pipeline.conf import settings
from pipeline.component_framework import context
from pipeline.utils.register import autodiscover_collections

logger = logging.getLogger("root")

DJANGO_MANAGE_CMD = "manage.py"
INIT_PASS_TRIGGER = {"migrate"}


class ComponentFrameworkConfig(AppConfig):
    name = "pipeline.component_framework"
    verbose_name = "PipelineComponentFramework"

    def ready(self):
        """
        @summary: 注册公共部分和当前RUN_VER下的标准插件到数据库
        @return:
        """

        if sys.argv and sys.argv[0] == DJANGO_MANAGE_CMD:
            try:
                command = sys.argv[1]
            except IndexError:
                return
            else:
                if command in INIT_PASS_TRIGGER:
                    print("ignore components init for command: {}".format(sys.argv))
                    return

        for path in settings.COMPONENT_AUTO_DISCOVER_PATH:
            autodiscover_collections(path)

        if context.skip_update_comp_models():
            return

        from pipeline.component_framework.models import ComponentModel
        from pipeline.component_framework.library import ComponentLibrary

        try:
            print("update component models")
            ComponentModel.objects.all().update(status=False)
            for code in ComponentLibrary.codes():
                ComponentModel.objects.filter(code=code, version__in=ComponentLibrary.versions(code)).update(
                    status=True
                )
            print("update component models finish")
        except InternalError as e:
            # version field migration
            logger.exception(e)
        except (ProgrammingError, OperationalError) as e:
            # first migrate
            logger.exception(e)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import imp
import sys

from django.test import TestCase

from pipeline.contrib.external_plugins.tests.mock import *  # noqa
from pipeline.contrib.external_plugins.tests.mock_settings import *  # noqa
from pipeline.contrib.external_plugins.utils.importer.base import NonstandardModuleImporter


class DummyImporter(NonstandardModuleImporter):
    def __init__(self, **kwargs):
        super(DummyImporter, self).__init__(modules=kwargs.get("modules", []))
        self._is_package = kwargs.get("is_package")
        self._get_code = kwargs.get("get_code")
        self._get_source = kwargs.get("get_source")
        self._get_file = kwargs.get("get_file")
        self._get_path = kwargs.get("get_path")

        self._accept_find_module_request_hook = MagicMock()
        self._pre_load_module_hook = MagicMock()
        self._post_load_module_hook = MagicMock()
        self._import_error_hook = MagicMock()

    def is_package(self, fullname):
        return self._is_package

    def get_code(self, fullname):
        return self._get_code

    def get_source(self, fullname):
        return self._get_source

    def get_file(self, fullname):
        return self._get_file

    def get_path(self, fullname):
        return self._get_path

    def accept_find_module_request_hook(self, fullname, path):
        self._accept_find_module_request_hook(fullname=fullname, path=path)

    def pre_load_module_hook(self, fullname, module):
        self._pre_load_module_hook(fullname=fullname, module=module)

    def post_load_module_hook(self, fullname, module):
        self._post_load_module_hook(fullname=fullname, module=module)

    def import_error_hook(self, fullname):
        self._import_error_hook(fullname=fullname)


class NonstandardModuleImporterTestCase(TestCase):
    def setUp(self):
        self.imp_acquire_lock_patcher = patch(IMP_ACQUIRE_LOCK, MagicMock())
        self.imp_release_lock_patcher = patch(IMP_RELEASE_LOCK, MagicMock())
        self.importer_exec_src_code_patcher = patch(UTILS_IMPORTER_BASE_EXECUTE_SRC_CODE, MagicMock())

        self.imp_acquire_lock_patcher.start()
        self.imp_release_lock_patcher.start()
        self.importer_exec_src_code_patcher.start()

    def tearDown(self):
        self.imp_acquire_lock_patcher.stop()
        self.imp_release_lock_patcher.stop()
        self.importer_exec_src_code_patcher.stop()

    def test_find_module__module_not_in_self_modules(self):
        importer = DummyImporter()

        self.assertIsNone(importer.find_module("django"))
        importer._accept_find_module_request_hook.assert_not_called()

        self.assertIsNone(importer.find_module("django.test"))
        importer._accept_find_module_request_hook.assert_not_called()

        self.assertIsNone(importer.find_module("django.test.utils"))
        importer._accept_find_module_request_hook.assert_not_called()

    def test_find_module__module_in_built_in(self):
        importer = DummyImporter()

        self.assertIsNone(importer.find_module("math"))
        importer._accept_find_module_request_hook.assert_not_called()

    def test_find_module__module_has_name_repetition(self):
        importer = DummyImporter(modules=["magic_module"])

        self.assertIsNone(importer.find_module("magic_module.magic_sub_module.magic_module"))
        importer._accept_find_module_request_hook.assert_not_called()

    def test_find_module__accept(self):
        importer = DummyImporter(modules=["magic_module"])

        fullname = "magic_module"
        self.assertIs(importer, importer.find_module(fullname))
        importer._accept_find_module_request_hook.assert_called_once_with(fullname=fullname, path=None)
        importer._accept_find_module_request_hook.reset_mock()

        fullname = "magic_module.magic_sub_module_1"
        self.assertIs(importer, importer.find_module(fullname))
        importer._accept_find_module_request_hook.assert_called_once_with(fullname=fullname, path=None)
        importer._accept_find_module_request_hook.reset_mock()

        fullname = "magic_module.magic_sub_module_1.magic_sub_module_2"
        self.assertIs(importer, importer.find_module(fullname))
        importer._accept_find_module_request_hook.assert_called_once_with(fullname=fullname, path=None)
        importer._accept_find_module_request_hook.reset_mock()

    def test_load_module__module_already_in_sys_modules(self):
        fullname = "exist_module"
        mod = Object()
        importer = DummyImporter()

        with patch(SYS_MODULES, {fullname: mod}):
            self.assertEqual(importer.load_module(fullname=fullname), mod)
            imp.acquire_lock.assert_called_once()
            imp.release_lock.assert_called_once()

    def test_load_module__get_source_raise_import_error(self):
        sub_module = "sub_module"
        fullname = "exist_module.sub_module"
        mod = Object()
        importer = DummyImporter()
        importer.get_source = MagicMock(side_effect=ImportError)

        with patch(SYS_MODULES, {sub_module: mod}):
            self.assertIsNone(importer.load_module(fullname=fullname))
            imp.acquire_lock.assert_called_once()
            imp.release_lock.assert_called_once()

    def test_load_module__is_package(self):
        src_code = "src_code"
        fullname = "magic_module"
        _file = "file"
        path = "path"
        importer = DummyImporter(is_package=True, get_source=src_code, get_file=_file, get_path=path)

        with patch(SYS_MODULES, {}):
            mod = importer.load_module(fullname=fullname)

            self.assertIs(sys.modules[fullname], mod)
            self.assertEqual(mod.__file__, _file)
            self.assertIs(mod.__loader__, importer)
            self.assertEqual(mod.__path__, path)
            self.assertEqual(mod.__package__, fullname)

            imp.acquire_lock.assert_called_once()
            importer._pre_load_module_hook.assert_called_once_with(fullname=fullname, module=mod)
            importer._execute_src_code.assert_called_once_with(src_code=src_code, module=mod)
            importer._post_load_module_hook.assert_called_once_with(fullname=fullname, module=mod)
            imp.release_lock.assert_called_once()

    def test_load_module__is_not_package(self):
        src_code = "src_code"
        fullname = "magic_module.sub_module"
        _file = "file"
        importer = DummyImporter(is_package=False, get_source=src_code, get_file=_file)

        with patch(SYS_MODULES, {}):
            mod = importer.load_module(fullname=fullname)

            self.assertIs(sys.modules[fullname], mod)
            self.assertEqual(mod.__file__, _file)
            self.assertIs(mod.__loader__, importer)
            self.assertEqual(mod.__package__, fullname.rpartition(".")[0])

            imp.acquire_lock.assert_called_once()
            importer._pre_load_module_hook.assert_called_once_with(fullname=fullname, module=mod)
            importer._execute_src_code.assert_called_once_with(src_code=src_code, module=mod)
            importer._post_load_module_hook.assert_called_once_with(fullname=fullname, module=mod)
            imp.release_lock.assert_called_once()

    def test_load_module__raise_exception_before_add_module(self):
        fullname = "magic_module.sub_module"
        importer = DummyImporter(is_package=False)
        importer.get_source = MagicMock(side_effect=Exception())
        importer._import_error_hook = MagicMock(side_effect=Exception())

        with patch(SYS_MODULES, {}):
            self.assertRaises(ImportError, importer.load_module, fullname)
            self.assertNotIn(fullname, sys.modules)

            importer._import_error_hook.assert_called_once()
            imp.release_lock.assert_called_once()

    def test_load_module__raise_exception_after_add_module(self):
        fullname = "magic_module.sub_module"
        importer = DummyImporter(is_package=False)
        importer.get_file = MagicMock(side_effect=Exception())

        with patch(SYS_MODULES, {}):
            self.assertRaises(ImportError, importer.load_module, fullname)
            self.assertNotIn(fullname, sys.modules)

            importer._import_error_hook.assert_called_once()
            imp.release_lock.assert_called_once()



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2021 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import contextlib
import logging
import traceback

from django.db import transaction

from pipeline.django_signal_valve import valve
from pipeline.engine.core import context
from pipeline.engine import exceptions, signals, states
from pipeline.engine.core.data import delete_parent_data, get_schedule_parent_data, set_schedule_data
from pipeline.engine.models import Data, MultiCallbackData, PipelineProcess, ScheduleService, Status

logger = logging.getLogger("pipeline_engine")


@contextlib.contextmanager
def schedule_exception_handler(process_id, schedule_id):
    try:
        yield
    except Exception as e:
        activity_id = schedule_id[: ScheduleService.SCHEDULE_ID_SPLIT_DIVISION]
        version = schedule_id[ScheduleService.SCHEDULE_ID_SPLIT_DIVISION :]
        if Status.objects.filter(id=activity_id, version=version).exists():
            logger.error(traceback.format_exc())
            process = PipelineProcess.objects.get(id=process_id)
            process.exit_gracefully(e)
        else:
            logger.warning("schedule({} - {}) forced exit.".format(activity_id, version))

        delete_parent_data(schedule_id)


@contextlib.contextmanager
def auto_release_schedule_lock(schedule_id):
    yield
    # release schedule lock before exit schedule
    ScheduleService.objects.filter(id=schedule_id, is_scheduling=True).update(is_scheduling=False)
    logger.warning("schedule({}) unlock success.".format(schedule_id))


def schedule(process_id, schedule_id, data_id=None):
    """
    调度服务主函数
    :param process_id: 被调度的节点所属的 PipelineProcess
    :param schedule_id: 调度 ID
    :param data_id: 回调数据ID
    :return:
    """
    with schedule_exception_handler(process_id, schedule_id):
        # set up context
        context.set_node_id(schedule_id[: ScheduleService.SCHEDULE_ID_SPLIT_DIVISION])

        # schedule maybe destroyed by other schedule
        try:
            sched_service = ScheduleService.objects.get(id=schedule_id)
            # stop if schedule status finished
            if sched_service.is_finished:
                logger.warning("schedule already finished, give up, sched_id: {}".format(schedule_id))
                return
        except ScheduleService.DoesNotExist:
            logger.warning("schedule not exist, give up, sched_id: {}".format(schedule_id))
            return

        # check whether the node is in a state waiting for scheduling
        service_act = sched_service.service_act
        act_id = sched_service.activity_id
        version = sched_service.version

        if not Status.objects.filter(id=act_id, version=version, state=states.RUNNING).exists():
            # forced failed
            logger.warning(
                "schedule service failed, schedule({} - {}) node state is not running or version do not match.".format(
                    act_id, version
                )
            )
            sched_service.destroy()
            return

        # try update lock schedule
        is_updated = ScheduleService.objects.filter(id=schedule_id, is_scheduling=False).update(is_scheduling=True)

        # lock failed, other worker may locking
        if is_updated == 0:
            # only retry at multi calback enabled case
            if not sched_service.multi_callback_enabled:
                logger.warning(
                    "invalid schedule request, schedule({} - {}) node state is not multi callback enabled type.".format(
                        act_id, version
                    )
                )
                return

            # retry lock after seconds
            logger.warning("schedule service lock-{} failed, retry after seconds".format(schedule_id))
            valve.send(
                signals,
                "schedule_ready",
                sender=ScheduleService,
                process_id=process_id,
                schedule_id=schedule_id,
                data_id=data_id,
                countdown=2,
            )
            return

        with auto_release_schedule_lock(schedule_id):
            # get data
            parent_data = get_schedule_parent_data(sched_service.id)
            if parent_data is None:
                raise exceptions.DataRetrieveError(
                    "child process({}) retrieve parent_data error, sched_id: {}".format(process_id, schedule_id)
                )

            # get schedule data
            if sched_service.multi_callback_enabled and data_id:
                try:
                    callback_data = MultiCallbackData.objects.get(id=data_id)
                    schedule_data = callback_data.data
                except MultiCallbackData.DoesNotExist:
                    logger.warning(
                        "schedule get callback_data failed, give up schedule, sched_id: {}".format(schedule_id)
                    )
                    return
            else:
                schedule_data = sched_service.callback_data

            # schedule
            ex_data, success = None, False
            try:
                success = service_act.schedule(parent_data, schedule_data)
                if success is None:
                    success = True
            except Exception:
                if service_act.error_ignorable:
                    success = True
                    service_act.ignore_error()
                    service_act.finish_schedule()

                ex_data = traceback.format_exc()
                logging.error(ex_data)

            sched_service.schedule_times += 1
            set_schedule_data(sched_service.id, parent_data)

            # schedule failed
            if not success:
                if not Status.objects.transit(id=act_id, version=version, to_state=states.FAILED).result:
                    # forced failed
                    logger.warning(
                        "FAILED transit failed, schedule({} - {}) had been forced exit.".format(act_id, version)
                    )
                    sched_service.destroy()
                    return

                if service_act.timeout:
                    signals.service_activity_timeout_monitor_end.send(
                        sender=service_act.__class__, node_id=service_act.id, version=version
                    )
                    logger.info("node {} {} timeout monitor revoke".format(service_act.id, version))

                Data.objects.write_node_data(service_act, ex_data=ex_data)

                with transaction.atomic():
                    process = PipelineProcess.objects.select_for_update().get(id=sched_service.process_id)
                    if not process.is_alive:
                        logger.info("pipeline %s has been revoked, status adjust failed." % process.root_pipeline_id)
                        return

                    process.adjust_status()

                # send activity error signal
                try:
                    service_act.schedule_fail()
                except Exception:
                    logger.error("schedule_fail handler fail: %s" % traceback.format_exc())

                signals.service_schedule_fail.send(
                    sender=ScheduleService, activity_shell=service_act, schedule_service=sched_service, ex_data=ex_data
                )

                valve.send(
                    signals,
                    "activity_failed",
                    sender=process.root_pipeline,
                    pipeline_id=process.root_pipeline_id,
                    pipeline_activity_id=service_act.id,
                    subprocess_id_stack=process.subprocess_stack,
                )
                return

            # schedule execute finished or one time callback finished
            if service_act.is_schedule_done() or sched_service.is_one_time_callback():
                error_ignorable = not service_act.get_result_bit()
                if not Status.objects.transit(id=act_id, version=version, to_state=states.FINISHED).result:
                    # forced failed
                    logger.warning(
                        "FINISHED transit failed, schedule({} - {}) had been forced exit.".format(act_id, version)
                    )
                    sched_service.destroy()
                    return

                if service_act.timeout:
                    signals.service_activity_timeout_monitor_end.send(
                        sender=service_act.__class__, node_id=service_act.id, version=version
                    )
                    logger.info("node {} {} timeout monitor revoke".format(service_act.id, version))

                Data.objects.write_node_data(service_act)
                if error_ignorable:
                    s = Status.objects.get(id=act_id)
                    s.error_ignorable = True
                    s.save()

                # sync parent data
                process = PipelineProcess.objects.get(id=sched_service.process_id)
                if not process.is_alive:
                    logger.warning("schedule({} - {}) revoked.".format(act_id, version))
                    sched_service.destroy()
                    return

                process.top_pipeline.data.update_outputs(parent_data.get_outputs())
                # extract outputs
                process.top_pipeline.context.extract_output(service_act)
                process.save(save_snapshot=True)

                # clear temp data
                delete_parent_data(sched_service.id)
                # save schedule service
                sched_service.finish()

                signals.service_schedule_success.send(
                    sender=ScheduleService, activity_shell=service_act, schedule_service=sched_service
                )

                valve.send(
                    signals,
                    "wake_from_schedule",
                    sender=ScheduleService,
                    process_id=sched_service.process_id,
                    activity_id=sched_service.activity_id,
                )
            else:
                Data.objects.write_node_data(service_act)
                if sched_service.multi_callback_enabled:
                    sched_service.save()
                else:
                    sched_service.set_next_schedule()



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017-2020 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import logging
import datetime
from dateutil.relativedelta import relativedelta
from celery import task
from celery.schedules import crontab
from celery.task import periodic_task
from django.db import transaction, connection

from pipeline.conf import default_settings
from pipeline.core.pipeline import Pipeline
from pipeline.engine import api, signals, states
from pipeline.engine.core import runtime, schedule
from pipeline.engine.health import zombie
from pipeline.engine.models import (
    NodeCeleryTask,
    NodeRelationship,
    PipelineProcess,
    ProcessCeleryTask,
    Status,
    ScheduleService,
    History,
)
from pipeline.models import PipelineInstance

logger = logging.getLogger("celery")


@task(ignore_result=True)
def process_unfreeze(process_id):
    process = PipelineProcess.objects.get(id=process_id)
    if not process.is_alive:
        logger.warning("process(%s) is not alive, mission cancel." % process_id)
        return

    runtime.run_loop(process)


@task(ignore_result=True)
def start(process_id):
    process = PipelineProcess.objects.get(id=process_id)
    if not process.is_alive:
        logger.warning("process(%s) is not alive, mission cancel." % process_id)
        return

    pipeline_id = process.root_pipeline.id
    # try to run
    action_result = Status.objects.transit(pipeline_id, states.RUNNING, is_pipeline=True, start=True)
    if not action_result.result:
        logger.warning("can not start pipeline({}), message: {}".format(pipeline_id, action_result.message))
        return

    NodeRelationship.objects.build_relationship(pipeline_id, pipeline_id)

    runtime.run_loop(process)


@task(ignore_result=True)
def dispatch(child_id):
    process = PipelineProcess.objects.get(id=child_id)
    if not process.is_alive:
        logger.info("process(%s) is not alive, mission cancel." % child_id)
        return

    runtime.run_loop(process)


@task(ignore_result=True)
def process_wake_up(process_id, current_node_id=None, call_from_child=False):
    process = PipelineProcess.objects.get(id=process_id)
    if not process.is_alive:
        logger.warning("process(%s) is not alive, mission cancel." % process_id)
        return

    pipeline_id = process.root_pipeline.id
    if not call_from_child:
        # success_when_unchanged to deal with parallel wake up
        action_result = Status.objects.transit(
            pipeline_id, to_state=states.RUNNING, is_pipeline=True, unchanged_pass=True
        )
        if not action_result.result:
            # BLOCKED is a tolerant running state
            if action_result.extra.state != states.BLOCKED:
                logger.warning("can not start pipeline({}), message: {}".format(pipeline_id, action_result.message))
                return

    process.wake_up()
    if current_node_id:
        process.current_node_id = current_node_id

    runtime.run_loop(process)


@task(ignore_result=True)
def wake_up(process_id):
    process = PipelineProcess.objects.get(id=process_id)
    if not process.is_alive:
        logger.warning("process(%s) is not alive, mission cancel." % process_id)
        return

    process.wake_up()
    runtime.run_loop(process)


@task(ignore_result=True)
def batch_wake_up(process_id_list, pipeline_id):
    action_result = Status.objects.transit(pipeline_id, to_state=states.RUNNING, is_pipeline=True)
    if not action_result.result:
        logger.warning("can not start pipeline({}), message: {}".format(pipeline_id, action_result.message))
        return
    for process_id in process_id_list:
        task_id = wake_up.apply_async(args=[process_id]).id
        ProcessCeleryTask.objects.bind(process_id, task_id)


@task(ignore_result=True)
def wake_from_schedule(process_id, service_act_id):
    process = PipelineProcess.objects.get(id=process_id)
    process.wake_up()

    service_act = process.top_pipeline.node(service_act_id)
    process.current_node_id = service_act.next().id
    runtime.run_loop(process)


@task(ignore_result=True)
def service_schedule(process_id, schedule_id, data_id=None):
    schedule.schedule(process_id, schedule_id, data_id)


@task(ignore_result=True)
def node_timeout_check(node_id, version, root_pipeline_id):
    NodeCeleryTask.objects.destroy(node_id)
    state = Status.objects.state_for(node_id, version=version, may_not_exist=True)
    if not state or state != states.RUNNING:
        logger.warning("node {} {} timeout kill failed, node not exist or not in running".format(node_id, version))
        return

    action_result = api.forced_fail(node_id, kill=True, ex_data="node execution timeout")
    if action_result.result:
        signals.activity_failed.send(sender=Pipeline, pipeline_id=root_pipeline_id, pipeline_activity_id=node_id)
    else:
        logger.warning("node {} - {} timeout kill failed".format(node_id, version))


@periodic_task(run_every=(crontab(**default_settings.ENGINE_ZOMBIE_PROCESS_HEAL_CRON)), ignore_result=True)
def heal_zombie_process():
    logger.info("Zombie process heal start")

    healer = zombie.get_healer()

    try:
        healer.heal()
    except Exception:
        logger.exception("An error occurred when healing zombies")

    logger.info("Zombie process heal finish")


@periodic_task(run_every=(crontab(**default_settings.EXPIRED_TASK_CLEAN_CRON)), ignore_result=True)
def expired_tasks_clean():
    if not default_settings.EXPIRED_TASK_CLEAN:
        logger.info("EXPIRED_TASK_CLEAN switch off, won't clean expired tasks.")
        return
    timestamp = datetime.datetime.now().timestamp()
    logger.info("Expired tasks clean start, timestamp: {}".format(timestamp))

    expired_create_time = datetime.date.today() - relativedelta(months=default_settings.TASK_EXPIRED_MONTH)
    pipeline_instance_ids = list(
        PipelineInstance.objects.filter(
            create_time__lte=expired_create_time, is_finished=True, is_revoked=False, is_expired=False
        )
        .order_by("create_time")
        .values_list("instance_id", flat=True)[: default_settings.EXPIRED_TASK_CLEAN_NUM_LIMIT]
    )
    logger.info(
        "Clean expired tasks before {} with tasks number: {}, instance ids: {}, timestamp: {}".format(
            expired_create_time, len(pipeline_instance_ids), ",".join(pipeline_instance_ids), timestamp
        )
    )

    for instance_id in pipeline_instance_ids:
        try:
            logger.info("Clean expired task: {}, timestamp: {}".format(instance_id, timestamp))
            _clean_pipeline_instance_data(instance_id, timestamp)
        except Exception as e:
            logger.exception(
                "An error occurred when clean expired task instance {}: {}, {}".format(instance_id, e, timestamp)
            )

    logger.info("Expired tasks clean finish, timestamp: {}".format(timestamp))


def _clean_pipeline_instance_data(instance_id, timestamp):
    """
    根据instance_id删除对应的任务数据
    """
    process_nodes = list(
        set(NodeRelationship.objects.filter(ancestor_id=instance_id).values_list("descendant_id", flat=True))
    )
    process_nodes = [process_node for process_node in process_nodes if process_node]
    process_nodes_regex = "^" + "|^".join(process_nodes) if process_nodes else ""
    pipeline_processes = PipelineProcess.objects.filter(root_pipeline_id=instance_id).values_list("id", "snapshot__id")
    pipeline_process_ids, process_snapshot_ids = [], []
    for process_id, snapshot_id in pipeline_processes:
        if process_id:
            pipeline_process_ids.append(process_id)
        if snapshot_id:
            process_snapshot_ids.append(snapshot_id)

    delete_subprocess_relationship = (
        "DELETE FROM `engine_subprocessrelationship` WHERE `engine_subprocessrelationship`.`process_id` IN (%s)"
    )
    delete_process_snapshot = "DELETE FROM `engine_processsnapshot` WHERE `engine_processsnapshot`.`id` IN (%s)"
    delete_pipeline_model = "DELETE FROM `engine_pipelinemodel` WHERE `engine_pipelinemodel`.`process_id` IN (%s)"
    delete_process_celery_task = (
        "DELETE FROM `engine_processcelerytask` WHERE `engine_processcelerytask`.`process_id` IN (%s)"
    )
    schedule_service_ids = list(
        ScheduleService.objects.filter(process_id__in=pipeline_process_ids).values_list("id", flat=True)
    )
    schedule_service_ids = [schedule_service_id for schedule_service_id in schedule_service_ids if schedule_service_id]
    delete_schedule_service = "DELETE FROM `engine_scheduleservice` WHERE `engine_scheduleservice`.`process_id` IN (%s)"
    delete_multi_callback_data = (
        "DELETE FROM `engine_multicallbackdata` WHERE `engine_multicallbackdata`.`schedule_id` IN (%s)"
    )
    delete_node_relationship = (
        "DELETE FROM `engine_noderelationship` "
        "WHERE (`engine_noderelationship`.`ancestor_id` IN (%s) "
        "OR `engine_noderelationship`.`descendant_id` IN (%s)) "
    )
    delete_node_celery_tasks = "DELETE FROM `engine_nodecelerytask` " "WHERE `engine_nodecelerytask`.`node_id` IN (%s)"
    delete_status = "DELETE FROM `engine_status` WHERE `engine_status`.`id` IN (%s)"
    delete_data = "DELETE FROM `engine_data` WHERE `engine_data`.`id` IN (%s)"
    delete_datasnapshot = "DELETE FROM `engine_datasnapshot` WHERE `engine_datasnapshot`.`key` REGEXP %s"
    delete_schedule_celery_task = (
        "DELETE FROM `engine_schedulecelerytask`" "WHERE `engine_schedulecelerytask`.`schedule_id` REGEXP %s"
    )
    history_data_ids = list(
        History.objects.filter(identifier__in=process_nodes).only("data").values_list("data__id", flat=True)
    )
    delete_history = "DELETE FROM `engine_history` WHERE `engine_history`.`identifier` IN (%s)"
    delete_history_data = "DELETE FROM `engine_historydata` WHERE `engine_historydata`.`id` IN (%s)"
    delete_pipeline_process = (
        "DELETE FROM `engine_pipelineprocess` " "WHERE `engine_pipelineprocess`.`root_pipeline_id` = %s"
    )
    with transaction.atomic():
        with connection.cursor() as cursor:
            if pipeline_process_ids:
                process_fs = _sql_format_strings(pipeline_process_ids)
                _raw_sql_execute(cursor, delete_subprocess_relationship % process_fs, pipeline_process_ids, timestamp)
                _raw_sql_execute(cursor, delete_pipeline_model % process_fs, pipeline_process_ids, timestamp)
                _raw_sql_execute(cursor, delete_process_celery_task % process_fs, pipeline_process_ids, timestamp)
                _raw_sql_execute(cursor, delete_schedule_service % process_fs, pipeline_process_ids, timestamp)
            if process_snapshot_ids:
                snapshot_fd = _sql_format_strings(process_snapshot_ids)
                _raw_sql_execute(cursor, delete_process_snapshot % snapshot_fd, process_snapshot_ids, timestamp)
            if schedule_service_ids:
                service_fd = _sql_format_strings(schedule_service_ids)
                _raw_sql_execute(cursor, delete_multi_callback_data % service_fd, schedule_service_ids, timestamp)
            if process_nodes:
                node_fs = _sql_format_strings(process_nodes)
                _raw_sql_execute(
                    cursor, delete_node_relationship % (node_fs, node_fs), process_nodes + process_nodes, timestamp
                )
                _raw_sql_execute(cursor, delete_node_celery_tasks % node_fs, process_nodes, timestamp)
                _raw_sql_execute(cursor, delete_status % node_fs, process_nodes, timestamp)
                _raw_sql_execute(cursor, delete_data % node_fs, process_nodes, timestamp)
                _raw_sql_execute(cursor, delete_history % node_fs, process_nodes, timestamp)
            if process_nodes_regex:
                _raw_sql_execute(cursor, delete_datasnapshot, [process_nodes_regex], timestamp)
                _raw_sql_execute(cursor, delete_schedule_celery_task, [process_nodes_regex], timestamp)
            if history_data_ids:
                history_fs = _sql_format_strings(history_data_ids)
                _raw_sql_execute(cursor, delete_history_data % history_fs, history_data_ids, timestamp)
            _raw_sql_execute(cursor, delete_pipeline_process, [instance_id], timestamp)
            PipelineInstance.objects.filter(instance_id=instance_id).update(is_expired=True)


def _sql_log(sql, params, timestamp):
    if isinstance(params, list):
        logger.info("[execute raw sql]: {}, timestamp: {}".format(sql % tuple(params), timestamp))
    else:
        logger.info("[execute raw sql]: {}, timestamp: {}".format(sql % params, timestamp))


def _sql_format_strings(list_data):
    return ",".join(["%s"] * len(list_data))


def _raw_sql_execute(cursor, sql, params, timestamp):
    _sql_log(sql, params, timestamp)
    cursor.execute(sql, params)
